
from .record import Record


class Split(Record):
    __attributes__ = ['left', 'delimiter', 'right', 'buffer']

    def __init__(self, left, delimiter, right, buffer=None):
        self.left = left
        self.delimiter = delimiter
        self.right = right
        self.buffer = buffer


class Splitter(Record):
    pass




import sys
import argparse
from random import seed, sample as sample_

from razdel import (
    sentenize,
    tokenize
)

from .partition import (
    parse_partitions,
    format_partitions,
    update_partitions
)
from .gen import (
    generate_partition_precision_tests,
    generate_partition_recall_tests
)
from .zoo import (
    dot_sentenize,
    deepmipt_sentenize,
    nltk_sentenize,
    segtok_sentenize,
    moses_sentenize,

    space_tokenize,
    re_tokenize,
    nltk_tokenize,
    spacy_tokenize,
    spacy_tokenize2,
    mystem_tokenize,
    moses_tokenize,
)


ZOO = {
    'dot_sentenize': dot_sentenize,
    'deepmipt_sentenize': deepmipt_sentenize,
    'nltk_sentenize': nltk_sentenize,
    'segtok_sentenize': segtok_sentenize,
    'moses_sentenize': moses_sentenize,
    'sentenize': sentenize,

    'space_tokenize': space_tokenize,
    're_tokenize': re_tokenize,
    'nltk_tokenize': nltk_tokenize,
    'spacy_tokenize': spacy_tokenize,
    'spacy_tokenize2': spacy_tokenize2,
    'mystem_tokenize': mystem_tokenize,
    'moses_tokenize': moses_tokenize,
    'tokenize': tokenize,
}


def stdin_lines():
    for line in sys.stdin:
        yield line.rstrip('\n')


def stdout_lines(lines):
    for line in lines:
        print(line)


def gen_(partitions, precision, recall):
    for partition in partitions:
        if precision:
            for test in generate_partition_precision_tests(partition):
                yield test
        if recall:
            for test in generate_partition_recall_tests(partition):
                yield test


def gen(args):
    precision = args.precision
    recall = args.recall
    if not precision and not recall:
        precision = True
        recall = True
    lines = stdin_lines()
    partitions = parse_partitions(lines)
    tests = gen_(partitions, precision, recall)
    lines = format_partitions(tests)
    stdout_lines(lines)


def sample(args):
    seed(args.seed)
    lines = list(stdin_lines())
    lines = sample_(lines, args.size)
    stdout_lines(lines)


def show_(guess, etalon):
    print('---etalon')
    for _ in etalon:
        print('>', _.text)
    print('---guess')
    for _ in guess:
        print('>', _.text)
    print()


def diff_(tests, segment, show):
    for test in tests:
        guess = list(segment(test.text))
        etalon = list(test.substrings)
        if guess != etalon:
            if show:
                show_(guess, etalon)
            else:
                yield test


def diff(args):
    segment = ZOO[args.segment]
    lines = stdin_lines()
    partitions = parse_partitions(lines)
    tests = diff_(partitions, segment, args.show)
    lines = format_partitions(tests)
    stdout_lines(lines)


def up(args):
    segment = ZOO[args.segment]
    lines = stdin_lines()
    partitions = parse_partitions(lines)
    partitions = update_partitions(partitions, segment)
    lines = format_partitions(partitions)
    stdout_lines(lines)


def main():
    parser = argparse.ArgumentParser(prog='razdel-ctl')
    parser.set_defaults(function=None)

    subs = parser.add_subparsers()

    sub = subs.add_parser('gen')
    sub.set_defaults(function=gen)
    sub.add_argument('--precision', action='store_true')
    sub.add_argument('--recall', action='store_true')

    sub = subs.add_parser('sample')
    sub.set_defaults(function=sample)
    sub.add_argument('size', type=int)
    sub.add_argument('--seed', type=int, default=1)

    sub = subs.add_parser('diff')
    sub.set_defaults(function=diff)
    sub.add_argument('segment', choices=ZOO)
    sub.add_argument('--show', action='store_true')

    sub = subs.add_parser('up')
    sub.set_defaults(function=up)
    sub.add_argument('segment', choices=ZOO)

    args = sys.argv[1:]
    args = parser.parse_args(args)
    if not args.function:
        parser.print_help()
        parser.exit()
    try:
        args.function(args)
    except BrokenPipeError:
        pass


if __name__ == '__main__':
    main()




from razdel.record import Record
from razdel.rule import JOIN
from razdel.substring import find_substrings


def safe_next(iter):
    try:
        return next(iter)
    except StopIteration:
        return


class Segmenter(Record):
    __attributes__ = ['split', 'rules']

    def __init__(self, split, rules):
        self.split = split
        self.rules = rules

    def join(self, split):
        for rule in self.rules:
            action = rule(split)
            if action:
                return action == JOIN

    def segment(self, parts):
        buffer = safe_next(parts)
        if buffer is None:
            return

        for split in parts:
            right = next(parts)
            split.buffer = buffer
            if self.join(split):
                buffer = buffer + split.delimiter + right
            else:
                yield buffer + split.delimiter
                buffer = right
        yield buffer

    post = None

    def __call__(self, text):
        parts = self.split(text)
        chunks = self.segment(parts)
        if self.post:
            chunks = self.post(chunks)
        return find_substrings(chunks, text)


class DebugSegmenter(Segmenter):
    def join(self, split):
        print(split.left, '|', split.delimiter, '|', split.right)
        for rule in self.rules:
            action = rule(split)
            if action:
                print('\t', action, rule.name)
                return action == JOIN





def parse_sokrs(lines):
    for line in lines:
        for word in line.split():
            yield word


def parse_pair_sokrs(lines):
    for line in lines:
        yield tuple(line.split())


TAIL_SOKRS = set(parse_sokrs([
    'дес тыс млн млрд',
    'дол долл',
    'коп руб р',
    'проц',  # 95 проц. акций,

    'га',
    'барр',  # 40 долларов за барр.
    'куб',  # 1000 куб. метр.
    'кв км',  # 700 тыс. кв. км.
    'см',  # 30 см

    'час мин сек',  # в 15 час. 13 мин. 53 сек.
    'в вв',  # XII в. XVIII—XIX вв.
    'г гг',  # 1996-1999гг

    'с стр',  # 287 стр.

    'co corp inc',

    'изд ed',  # 1-е изд. Arthur W. Hummel, ed. Eminent Chinese

    'др',  # и другие
    'al',  # North et al.
]))

HEAD_SOKRS = set(parse_sokrs([
    'букв',  # яп. 18禁, букв. «запрещено
    'ст',  # ст.-слав.
    'трад',  # кит. трад
    'лат венг исп кат укр нем англ фр итал греч',
    'евр араб яп слав кит рус русск латв',
    'словацк хорв',

    'mr mrs ms dr vs',
    'св',  # св.Иоанна
    'арх зав зам проф акад',
    'кн',  # кандидат наук
    'корр',  # сообщил корр. ИТАР-ТАСС
    'ред',  # Под ред. Линды Уильямс
    'гр',  # гр. Валевской
    'ср',  # Ср. L. Ross
    'чл корр',  # является чл.-корр. АН СССР
    'им',  # им. Вс. Мейерхольда
    'тов',  # тюремном подвале тов. Берия

    'нач пол',  # нач. XX века

    'chap',
    'п пп ст ч чч гл стр абз пт',  # ст. 129 ч. 2 п. 8 Гл. VI
    'no',  # No. 6

    'просп пр ул ш г гор д стр к корп пер корп обл эт пом ауд оф ком комн каб',
    'домовлад лит',
    'т',  # т. 1 л.д. 85-89
    'рп пос с х',  # х. Ново-Максимовский, с. Кляшево рп.Раздолинск
    'пл',  # площадь
    'bd',  # Bd. 16, Berlin
    'о оз',  # Вблизи оз. Селяха
    'р',  # р. Иордан
    'а',  # а. Адыге-Хабль

    'обр',  # обр. 1936 г.
    'ум',  # ум. 1064
    'ок',  # "родилась ок. 1211", "работают ок. 150 специалистов"
    'откр',  # Откр. 20:40

    'пс ps',
    'upd',
    'см',
    'напр',  # UNIX-семейства, напр. Linux, FreeBSD
    'доп',

    'юр физ',  # юр. адрес

    'тел',
    'сб',  # Сб. «Киноварь»
    'внутр',  # к внутр. миру героев
    'дифф',  # мне по дифф. зачёту «5» поставил
    'гос',  # гос. экзамены
    'отм',  # от отм. 0.000
]))

OTHER_SOKRS = set(parse_sokrs([
    'сокр рис искл прим',

    'яз',
    'устар',  # пометкой "устар."
    'шутл',  # "в стиле шутл.", "bones — шутл. человек"
]))

SOKRS = TAIL_SOKRS | HEAD_SOKRS | OTHER_SOKRS

TAIL_PAIR_SOKRS = set(parse_pair_sokrs([
    'т п',
    'т д',
    'у е',
    'н э',
    'p m',
    'a m',
    'с г',  # от 18 мая с. г.
    'р х',  # 250 года до Р. Х.
    'с г',  # 18 мая с. г.
    'с ш',  # 50°13′ с. ш.
    'з д',  # 12°48′ з. д.
    'л с',
    'ч т', 'т д',  # ч.т.д
]))

HEAD_PAIR_SOKRS = set(parse_pair_sokrs([
    'т е',
    'т к',
    'т н',
    'и о',
    'к н',
    'к п', 'п н',  # к.п.н
    'к т', 'т н',  # к.т.н
    'л д',  # т. 1 л.д. 85-89
]))

OTHER_PAIR_SOKRS = set(parse_pair_sokrs([
    'ед ч',
    'мн ч',
    'повел накл',  # в 1 лице мн. ч. повел. накл.
    'жен р'
    'муж р',
]))

PAIR_SOKRS = TAIL_PAIR_SOKRS | HEAD_PAIR_SOKRS | OTHER_PAIR_SOKRS

INITIALS = {
    'дж',
    'ed',
    'вс',  # Вс. Мейерхольда
}




from .partition import Partition


def generate_partition_precision_tests(partition):
    for substring in partition.substrings:
        yield Partition([substring.text])


def generate_precision_tests(partitions):
    for partition in partitions:
        for test in generate_partition_precision_tests(partition):
            yield test

def correct_precision(segment, partition):
    etalon = list(partition.substrings)
    assert len(etalon) == 1
    guess = list(segment(partition.text))
    return len(guess) == 1


def generate_partition_recall_tests(partition):
    is_fill = partition.is_fill
    chunks = partition.chunks
    for index in range(1, len(chunks) - 1):
        previous = chunks[index - 1]
        current = chunks[index]
        next = chunks[index + 1]
        if is_fill(current) and not is_fill(previous) and not is_fill(next):
            yield Partition([previous, current, next])


def generate_recall_tests(partitions):
    for partition in partitions:
        for test in generate_partition_recall_tests(partition):
            yield test


def correct_recall(segment, partition):
    substrings = list(partition.substrings)
    assert len(substrings) == 2
    etalon = substrings[0].stop
    for part in segment(partition.text):
        if part.stop == etalon:
            return True
    return False





class cached_property(object):
    def __init__(self, function):
        self.function = function
        self.name = function.__name__

    def __get__(self, instance, type=None):
        if self.name not in instance.__dict__:
            result = instance.__dict__[self.name] = self.function(instance)
            return result
        return instance.__dict__[self.name]


class Record(object):
    __attributes__ = []

    def __eq__(self, other):
        return (
            type(self) == type(other)
            and all(
                (getattr(self, _) == getattr(other, _))
                for _ in self.__attributes__
            )
        )

    def __ne__(self, other):
        return not self == other

    def __iter__(self):
        return (getattr(self, _) for _ in self.__attributes__)

    def __hash__(self):
        return hash(tuple(self))

    def __repr__(self):
        name = self.__class__.__name__
        args = ', '.join(
            repr(getattr(self, _))
            for _ in self.__attributes__
        )
        return '{name}({args})'.format(
            name=name,
            args=args
        )

    def _repr_pretty_(self, printer, cycle):
        name = self.__class__.__name__
        if cycle:
            printer.text('{name}(...)'.format(name=name))
        else:
            with printer.group(len(name) + 1, '{name}('.format(name=name), ')'):
                for index, key in enumerate(self.__attributes__):
                    if index > 0:
                        printer.text(',')
                        printer.breakable()
                    value = getattr(self, key)
                    printer.pretty(value)




import re
import logging
from contextlib import contextmanager

from razdel.substring import find_substrings


def dot_sentenize_(text):
    previous = 0
    for match in re.finditer(r'([.?!…])\s+', text, re.U):
        delimiter = match.group(1)
        start = match.start()
        yield text[previous:start] + delimiter
        previous = match.end()
    if previous < len(text):
        yield text[previous:]


def dot_sentenize(text):
    chunks = dot_sentenize_(text)
    return find_substrings(chunks, text)


@contextmanager
def no_logger(logger):
    logger.disabled = True
    try:
        yield
    finally:
        logger.disabled = False


LOGGER = logging.getLogger()


def deepmipt_sentenize(text):
    from rusenttokenize import ru_sent_tokenize

    with no_logger(LOGGER):
        chunks = ru_sent_tokenize(text)
    return find_substrings(chunks, text)


def nltk_sentenize(text):
    from nltk import sent_tokenize

    chunks = sent_tokenize(text, 'russian')
    return find_substrings(chunks, text)


def segtok_sentenize(text):
    from segtok.segmenter import split_single

    chunks = split_single(text)
    return find_substrings(chunks, text)


MOSES_SENT = None


def moses_sentenize(text):
    from mosestokenizer import MosesSentenceSplitter

    global MOSES_SENT
    if not MOSES_SENT:
        MOSES_SENT = MosesSentenceSplitter('ru')

    chunks = MOSES_SENT([text])
    return find_substrings(chunks, text)


TOKEN = re.compile(r'([^\W\d]+|\d+|[^\w\s])', re.U)


def space_tokenize(text):
    chunks = re.split(r'\s+', text)
    return find_substrings(chunks, text)


def re_tokenize(text):
    chunks = TOKEN.findall(text)
    return find_substrings(chunks, text)


def nltk_tokenize(text):
    from nltk.tokenize import word_tokenize

    chunks = word_tokenize(text, 'russian')
    return find_substrings(chunks, text)


NLP = None


def spacy_tokenize(text):
    from spacy.lang.ru import Russian

    global NLP
    if not NLP:
        NLP = Russian()

    doc = NLP(text)
    chunks = [token.text for token in doc]
    return find_substrings(chunks, text)


NLP2 = None


def spacy_tokenize2(text):
    from spacy.lang.ru import Russian
    from spacy_russian_tokenizer import (
        RussianTokenizer,
        MERGE_PATTERNS,
        SYNTAGRUS_RARE_CASES
    )

    global NLP2
    if not NLP2:
        NLP2 = Russian()
        NLP2.add_pipe(
            RussianTokenizer(NLP2, MERGE_PATTERNS + SYNTAGRUS_RARE_CASES),
            name='russian_tokenizer'
        )

    doc = NLP2(text)
    chunks = [token.text for token in doc]
    return find_substrings(chunks, text)


# {'analysis': [], 'text': 'Mystem'},
#  {'text': ' — '},
#  {'analysis': [{'lex': 'консольный'}], 'text': 'консольная'},
#  {'text': ' '},
#  {'analysis': [{'lex': 'программа'}], 'text': 'программа'},
#  {'text': '\n'}]


MYSTEM = None


def parse_mystem(data):
    for item in data:
        text = item['text'].strip()
        if text:
            yield text


def mystem_tokenize(text):
    from pymystem3 import Mystem

    global MYSTEM
    if not MYSTEM:
        MYSTEM = Mystem(
            grammar_info=False,
            entire_input=True,
            disambiguation=False,
            weight=False
        )

    data = MYSTEM.analyze(text)
    chunks = parse_mystem(data)
    return find_substrings(chunks, text)


MOSES_TOK = None


def moses_tokenize(text):
    from mosestokenizer import MosesTokenizer

    global MOSES_TOK
    if not MOSES_TOK:
        MOSES_TOK = MosesTokenizer('ru')
        # disable
        MOSES_TOK.argv.append('-no-escape')  # " -> &quot;
        MOSES_TOK.argv.remove('-a')  # - -> @-@
        MOSES_TOK.restart()

    chunks = MOSES_TOK(text)
    return find_substrings(chunks, text)


def segtok_tokenize(text):
    from segtok.tokenizer import word_tokenizer

    chunks = word_tokenizer(text)
    return find_substrings(chunks, text)




import re

from razdel.record import (
    Record,
    cached_property
)
from razdel.rule import (
    JOIN, SPLIT,
    Rule,
    FunctionRule
)
from razdel.split import (
    Split,
    Splitter,
)

from .base import (
    safe_next,
    Segmenter,
    DebugSegmenter
)
from .punct import (
    DASHES,
    ENDINGS,
    QUOTES,
    BRACKETS,

    SMILES
)


RU = 'RU'
LAT = 'LAT'
INT = 'INT'
PUNCT = 'PUNCT'
OTHER = 'OTHER'

PUNCTS = '\\/!#$%&*+,.:;<=>?@^_`|~№…' + DASHES + QUOTES + BRACKETS

ATOM = re.compile(
    r'''
    (?P<RU>[а-яё]+)
    |(?P<LAT>[a-z]+)
    |(?P<INT>\d+)
    |(?P<PUNCT>[%s])
    |(?P<OTHER>\S)
    ''' % re.escape(PUNCTS),
    re.I | re.U | re.X
)

SMILE = re.compile(r'^' + SMILES + '$', re.U)


##########
#
#  TRIVIAL
#
######


def split_space(split):
    if split.delimiter:
        return SPLIT


##########
#
#   2112
#
##########


class Rule2112(Rule):
    def __call__(self, split):
        if self.delimiter(split.left):
            # cho-|to
            left, right = split.left_2, split.right_1
        elif self.delimiter(split.right):
            # cho|-to
            left, right = split.left_1, split.right_2
        else:
            return

        if not left or not right:
            return

        return self.rule(left, right)


class DashRule(Rule2112):
    name = 'dash'

    def delimiter(self, delimiter):
        return delimiter in DASHES

    def rule(self, left, right):
        if left.type == PUNCT or right.type == PUNCT:
            return
        return JOIN


class UnderscoreRule(Rule2112):
    name = 'underscore'

    def delimiter(self, delimiter):
        return delimiter == '_'

    def rule(self, left, right):
        if left.type == PUNCT or right.type == PUNCT:
            return
        return JOIN


class FloatRule(Rule2112):
    name = 'float'

    def delimiter(self, delimiter):
        return delimiter in '.,'

    def rule(self, left, right):
        if left.type == INT and right.type == INT:
            return JOIN


class FractionRule(Rule2112):
    name = 'fraction'

    def delimiter(self, delimiter):
        return delimiter in '/\\'

    def rule(self, left, right):
        if left.type == INT and right.type == INT:
            return JOIN


########
#
#   PUNCT
#
##########


def punct(split):
    if split.left_1.type != PUNCT or split.right_1.type != PUNCT:
        return

    left = split.left
    right = split.right

    if SMILE.match(split.buffer + right):
        return JOIN

    if left in ENDINGS and right in ENDINGS:
        # ... ?!
        return JOIN

    if left + right in ('--', '**'):
        # ***
        return JOIN


def other(split):
    left = split.left_1.type
    right = split.right_1.type

    if left == OTHER and right in (OTHER, RU, LAT):
        # ΔP
        return JOIN

    if left in (OTHER, RU, LAT) and right == OTHER:
        # Δσ mβж
        return JOIN


########
#
#  EXCEPTION
#
#######


def yahoo(split):
    if split.left_1.normal == 'yahoo' and split.right == '!':
        return JOIN


########
#
#   SPLIT
#
########


class Atom(Record):
    __attributes__ = ['start', 'stop', 'type', 'text']

    def __init__(self, start, stop, type, text):
        self.start = start
        self.stop = stop
        self.type = type
        self.text = text
        self.normal = text.lower()


class TokenSplit(Split):
    def __init__(self, left, delimiter, right):
        self.left_atoms = left
        self.right_atoms = right
        super(TokenSplit, self).__init__(
            self.left_1.text,
            delimiter,
            self.right_1.text
        )

    @cached_property
    def left_1(self):
        return self.left_atoms[-1]

    @cached_property
    def left_2(self):
        if len(self.left_atoms) > 1:
            return self.left_atoms[-2]

    @cached_property
    def left_3(self):
        if len(self.left_atoms) > 2:
            return self.left_atoms[-3]

    @cached_property
    def right_1(self):
        return self.right_atoms[0]

    @cached_property
    def right_2(self):
        if len(self.right_atoms) > 1:
            return self.right_atoms[1]

    @cached_property
    def right_3(self):
        if len(self.right_atoms) > 2:
            return self.right_atoms[2]


class TokenSplitter(Splitter):
    def __init__(self, window=3):
        self.window = window

    def atoms(self, text):
        matches = ATOM.finditer(text)
        for match in matches:
            start = match.start()
            stop = match.end()
            type = match.lastgroup
            text = match.group(0)
            yield Atom(
                start, stop,
                type, text
            )

    def __call__(self, text):
        atoms = list(self.atoms(text))
        for index in range(len(atoms)):
            atom = atoms[index]
            if index > 0:
                previous = atoms[index - 1]
                delimiter = text[previous.stop:atom.start]
                left = atoms[max(0, index - self.window):index]
                right = atoms[index:index + self.window]
                yield TokenSplit(left, delimiter, right)
            yield atom.text


########
#
#   SEGMENT
#
########


RULES = [
    DashRule(),
    UnderscoreRule(),
    FloatRule(),
    FractionRule(),

    FunctionRule(punct),
    FunctionRule(other),

    FunctionRule(yahoo),
]


class TokenSegmenter(Segmenter):
    def __init__(self):
        super(TokenSegmenter, self).__init__(TokenSplitter(), RULES)

    def segment(self, parts):
        buffer = safe_next(parts)
        if buffer is None:
            return

        for split in parts:
            right = next(parts)
            split.buffer = buffer
            if not split.delimiter and self.join(split):
                buffer += right
            else:
                yield buffer
                buffer = right
        yield buffer

    @property
    def debug(self):
        return DebugTokenSegmenter()


class DebugTokenSegmenter(TokenSegmenter, DebugSegmenter):
    pass


tokenize = TokenSegmenter()





def pytest_assertrepr_compare(op, left, right):
    # add one more line for "assert ..."
    return (
        ['']
        + ['---']
        + ['> ' + _.text for _ in left]
        + ['---']
        + ['> ' + _.text for _ in right]
    )


def pytest_addoption(parser):
    parser.addoption('--int', type='int')


def pytest_generate_tests(metafunc):
    if 'int_test' in metafunc.fixturenames:
        tests = []
        count = metafunc.config.getoption('int')
        if count:
            tests = metafunc.module.int_tests(count)
        metafunc.parametrize('int_test', tests)




import re

from razdel.record import Record
from razdel.substring import Substring


FILL = ' '
FILL_PATTERN = re.compile('^\s*$')


class Partition(Record):
    __attributes__ = ['chunks']

    is_fill = FILL_PATTERN.match

    def __init__(self, chunks):
        self.chunks = chunks

    @property
    def text(self):
        return ''.join(self.chunks)

    @property
    def substrings(self):
        start = 0
        for chunk in self.chunks:
            stop = start + len(chunk)
            if not self.is_fill(chunk):
                yield Substring(start, stop, chunk)
            start = stop

    @classmethod
    def from_substrings(cls, substrings):
        chunks = list(substring_chunks(substrings))
        return cls(chunks)


def substring_chunks(substrings, fill=FILL):
    previous = 0
    for index, substring in enumerate(substrings):
        if index > 0:
            size = substring.start - previous
            if size:
                yield fill * size
        yield substring.text
        previous = substring.stop


SEP = '|'
ESCAPE = [
    (SEP, r'\|'),
    ('\n', r'\n')
]


def escape_chunk(chunk):
    for source, target in ESCAPE:
        chunk = chunk.replace(source, target)
    return chunk


def parse_partition(line):
    chunks = line.split(SEP)
    return Partition(chunks)


def parse_partitions(lines):
    for line in lines:
        yield parse_partition(line)


def format_partition(partition):
    return SEP.join(escape_chunk(_) for _ in partition.chunks)


def format_partitions(partitions):
    for partition in partitions:
        yield format_partition(partition)


def update_partition(partition, segment):
    text = partition.text
    substrings = segment(text)
    return Partition.from_substrings(substrings)


def update_partitions(partitions, segment):
    for partition in partitions:
        yield update_partition(partition, segment)




import pytest

from razdel import sentenize

from .partition import parse_partitions
from .common import (
    run,
    data_path,
    data_lines
)


UNIT = parse_partitions([
    'фонетических правил языка; в случае, если',
    '(Прилепин — очень хороший писатель, лучше, чем Лимонов.| |Но враг)',
    'Петров - 176!| |Михайлов - 180!',
    'если бы… не тот широко',
    'Георгий Иванов.| |На грани музыки и сна',
    'исполняется 150 лет.| |31 мая 1859 года после неоднократных',

    # SOKR
    'И т. д. и т. п.| |В общем, вся газета',
    'специалистом, к.п.н. И. П. Карташовым.',
    'основании п. 2, ст. 5 УПК',
    'Вблизи оз. Селяха',
    'уменьшить с 20 до 18 проц. (при сохранении',
    '6 июля 2007 г. "в связи с совершением',
    'на 500 тыс. машин',
    'Влияние взглядов Л. В. Щербы',
    'директор фирмы Чарльз Дж. Филлипс',
    'Т.е. ОБЯЗАТЕЛЬНО письменно',
    'была утечка т.н. Таблицы боевых действий',
    'В 1996-1999гг. теффт',
    'России, т. е. 55 % опрошенных',
    'я ощущал в 1990-е.| |Славное было время',

    # BOUND
    'словам, "не будет точно".| |"Возможно, у нас',
    'Брось!.."| |Связываться не хотелось',
    'Peter Goldreich,Scott Tremaine (1979).| |«Относительно теории колец Урана».',
    'Это чудовищные риски.| |"Яндекс" попал под удар',
    'кто они такие… »',

    # DASH
    '- "Так в чем же дело?"| |- "Не ра-ду-ют".',
    '— Ты ей скажи, что я ей гостинца дам.| |— А мне дашь?',

    '4. Я присутствовал во время встречи',
    'IV. Гестационный сахарный диабет',
    '§2. Нахождение оптимального объекта.',
    '8.1. Зачем нужны эти классы?',
    'в данной квартире;| |2) отчуждать свою долю',

    'пастухов - тоже ;)| |Я вспомнила',
    'распределённой жабы :))| |А платить мне будут аж 1200 рублей'
])


@pytest.mark.parametrize('test', UNIT)
def test_unit(test):
    run(sentenize, test)


def int_tests(count):
    path = data_path('sents.txt')
    lines = data_lines(path, count)
    return parse_partitions(lines)


def test_int(int_test):
    run(sentenize, int_test)




from .record import Record


class Substring(Record):
    __attributes__ = ['start', 'stop', 'text']

    def __init__(self, start, stop, text):
        self.start = start
        self.stop = stop
        self.text = text


def find_substrings(chunks, text):
    offset = 0
    for chunk in chunks:
        start = text.find(chunk, offset)
        stop = start + len(chunk)
        yield Substring(start, stop, chunk)
        offset = stop




from setuptools import setup, find_packages


with open('README.md') as file:
    description = file.read()


setup(
    name='razdel',
    version='0.5.0',

    description='Splits russian text into tokens, sentences, section. Rule-based',
    long_description=description,
    long_description_content_type='text/markdown',

    url='https://github.com/natasha/razdel',
    author='Alexander Kukushkin',
    author_email='alex@alexkuk.ru',
    license='MIT',

    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    keywords='nlp, natural language processing, russian, token, sentence, tokenize',

    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'razdel-ctl=razdel.tests.ctl:main'
        ],
    },
    install_requires=[]
)




from .record import Record


SPLIT = 'split'
JOIN = 'join'


class Rule(Record):
    name = None

    def __call__(self, split):
        raise NotImplementedError


class FunctionRule(Rule):
    __attributes__ = ['name']

    def __init__(self, function):
        self.name = function.__name__
        self.function = function

    def __call__(self, split):
        return self.function(split)




import re

from razdel.record import cached_property
from razdel.rule import (
    JOIN,
    FunctionRule
)
from razdel.split import (
    Split,
    Splitter,
)

from .base import (
    Segmenter,
    DebugSegmenter
)
from .sokr import (
    HEAD_SOKRS,
    SOKRS,

    HEAD_PAIR_SOKRS,
    PAIR_SOKRS,

    INITIALS
)
from .punct import (
    ENDINGS,
    DASHES,

    QUOTES,
    CLOSE_QUOTES,
    GENERIC_QUOTES,
    CLOSE_BRACKETS,

    SMILES,
)


SPACE_SUFFIX = re.compile(r'\s$', re.U)
SPACE_PREFIX = re.compile(r'^\s', re.U)

TOKEN = re.compile(r'([^\W\d]+|\d+|[^\w\s])', re.U)
FIRST_TOKEN = re.compile(r'^\s*([^\W\d]+|\d+|[^\w\s])', re.U)
LAST_TOKEN = re.compile(r'([^\W\d]+|\d+|[^\w\s])\s*$', re.U)
WORD = re.compile(r'([^\W\d]+|\d+)', re.U)
PAIR_SOKR = re.compile(r'(\w)\s*\.\s*(\w)\s*$', re.U)
INT_SOKR = re.compile(r'\d+\s*-?\s*(\w+)\s*$', re.U)

ROMAN = re.compile(r'^[IVXML]+$', re.U)
BULLET_CHARS = set('§абвгдеabcdef')
BULLET_BOUNDS = '.)'
BULLET_SIZE = 20

DELIMITERS = ENDINGS + ';' + GENERIC_QUOTES + CLOSE_QUOTES + CLOSE_BRACKETS
SMILE_PREFIX = re.compile(r'^\s*' + SMILES, re.U)


def is_lower_alpha(token):
    return token.isalpha() and token.islower()


#######
#
#   TRIVIAL
#
######


def empty_side(split):
    if not split.left_token or not split.right_token:
        return JOIN


def no_space_prefix(split):
    if not split.right_space_prefix:
        return JOIN


def lower_right(split):
    if is_lower_alpha(split.right_token):
        return JOIN


def delimiter_right(split):
    right = split.right_token
    if right in GENERIC_QUOTES:
        return

    if right in DELIMITERS:
        return JOIN

    if SMILE_PREFIX.match(split.right):
        return JOIN


########
#
#   SOKR
#
#########


def is_sokr(token):
    if token.isdigit():
        return True
    if not token.isalpha():
        return True  # punct
    return token.islower()  # lower alpha


def sokr_left(split):
    if split.delimiter != '.':
        return

    right = split.right_token
    match = split.left_pair_sokr
    if match:
        a, b = match
        left = a.lower(), b.lower()

        if left in HEAD_PAIR_SOKRS:
            return JOIN

        if left in PAIR_SOKRS:
            if is_sokr(right):
                return JOIN
            return

    left = split.left_token.lower()
    if left in HEAD_SOKRS:
        return JOIN

    if left in SOKRS and is_sokr(right):
            return JOIN


def inside_pair_sokr(split):
    if split.delimiter != '.':
        return

    left = split.left_token.lower()
    right = split.right_token.lower()
    if (left, right) in PAIR_SOKRS:
        return JOIN


def initials_left(split):
    if split.delimiter != '.':
        return

    left = split.left_token
    if left.isupper() and len(left) == 1:
        return JOIN
    if left.lower() in INITIALS:
        return JOIN


##########
#
#  BOUND
#
########


def close_bound(split):
    left = split.left_token
    if left in ENDINGS:
        return
    return JOIN


def close_quote(split):
    delimiter = split.delimiter
    if delimiter not in QUOTES:
        return

    if delimiter in CLOSE_QUOTES:
        return close_bound(split)

    if delimiter in GENERIC_QUOTES:
        if not split.left_space_suffix:
            return close_bound(split)
        return JOIN


def close_bracket(split):
    if split.delimiter in CLOSE_BRACKETS:
        return close_bound(split)


#######
#
#   BULLET
#
########


def is_bullet(token):
    if token.isdigit():
        return True
    if token in BULLET_BOUNDS:
        # "8.1." "2)."
        return True
    if token.lower() in BULLET_CHARS:
        return True
    if ROMAN.match(token):
        return True


def list_item(split):
    if split.delimiter not in BULLET_BOUNDS:
        return

    if len(split.buffer) > BULLET_SIZE:
        return

    if all(is_bullet(_) for _ in split.buffer_tokens):
        return JOIN


#########
#
#   DASH
#
########


def dash_right(split):
    if split.right_token not in DASHES:
        return

    right = split.right_word
    if right and is_lower_alpha(right):
        return JOIN


##########
#
#   SPLIT
#
##########


class SentSplit(Split):
    @cached_property
    def right_space_prefix(self):
        return bool(SPACE_PREFIX.match(self.right))

    @cached_property
    def left_space_suffix(self):
        return bool(SPACE_SUFFIX.search(self.left))

    @cached_property
    def right_token(self):
        match = FIRST_TOKEN.match(self.right)
        if match:
            return match.group(1)

    @cached_property
    def left_token(self):
        match = LAST_TOKEN.search(self.left)
        if match:
            return match.group(1)

    @cached_property
    def left_pair_sokr(self):
        match = PAIR_SOKR.search(self.left)
        if match:
            return match.groups()

    @cached_property
    def left_int_sokr(self):
        match = INT_SOKR.search(self.left)
        if match:
            return match.group(1)

    @cached_property
    def right_word(self):
        match = WORD.search(self.right)
        if match:
            return match.group(1)

    @cached_property
    def buffer_tokens(self):
        return TOKEN.findall(self.buffer)

    @cached_property
    def buffer_first_token(self):
        match = FIRST_TOKEN.match(self.buffer)
        if match:
            return match.group(1)


DELIMITER = '({smiles}|[{delimiters}])'.format(
    delimiters=re.escape(DELIMITERS),
    smiles=SMILES
)


class SentSplitter(Splitter):
    __attributes__ = ['pattern', 'window']

    def __init__(self, pattern=DELIMITER, window=10):
        self.pattern = pattern
        self.window = window
        self.re = re.compile(pattern, re.U)

    def __call__(self, text):
        matches = self.re.finditer(text)
        previous = 0
        for match in matches:
            start = match.start()
            stop = match.end()
            delimiter = match.group(1)
            yield text[previous:start]
            left = text[max(0, start - self.window):start]
            right = text[stop:stop + self.window]
            yield SentSplit(left, delimiter, right)
            previous = stop
        yield text[previous:]


########
#
#   SEGMENT
#
########


RULES = [FunctionRule(_) for _ in [
    empty_side,
    no_space_prefix,
    lower_right,
    delimiter_right,

    sokr_left,
    inside_pair_sokr,
    initials_left,

    list_item,

    close_quote,
    close_bracket,

    dash_right,
]]


class SentSegmenter(Segmenter):
    def __init__(self, split=SentSplitter(), rules=RULES):
        super(SentSegmenter, self).__init__(split, rules)

    @property
    def debug(self):
        return DebugSentSegmenter()

    def post(self, chunks):
        for chunk in chunks:
            yield chunk.strip()


class DebugSentSegmenter(SentSegmenter, DebugSegmenter):
    pass


sentenize = SentSegmenter()





ENDINGS = '.?!…'
DASHES = '‑–—−-'

OPEN_QUOTES = '«“‘'
CLOSE_QUOTES = '»”’'
GENERIC_QUOTES = '"„\''
QUOTES = OPEN_QUOTES + CLOSE_QUOTES + GENERIC_QUOTES

OPEN_BRACKETS = '([}'
CLOSE_BRACKETS = ')]}'
BRACKETS = OPEN_BRACKETS + CLOSE_BRACKETS

BOUNDS = QUOTES + BRACKETS

SMILES = r'[=:;]-?[)(]{1,3}'  # :-) ;) =(((




import pytest

from razdel import tokenize

from .partition import parse_partitions
from .common import (
    run,
    data_path,
    data_lines
)


UNIT = parse_partitions([
    '1',
    'что-то',
    'К_тому_же',
    '...',
    '1,5',
    '1/2',

    '»||.',
    ')||.',
    '(||«',
    ':)))',
    ':)||,',

    'mβж',
    'Δσ',
    '',
])


@pytest.mark.parametrize('test', UNIT)
def test_unit(test):
    run(tokenize, test)


def int_tests(count):
    path = data_path('tokens.txt')
    lines = data_lines(path, count)
    return parse_partitions(lines)


def test_int(int_test):
    run(tokenize, int_test)




import os
from random import seed, sample


def run(segment, test):
    guess = list(segment(test.text))
    etalon = list(test.substrings)
    assert guess == etalon


def data_path(filename):
    return os.path.join(
        os.path.dirname(__file__),
        'data',
        filename
    )


def load_lines(path):
    with open(path) as file:
        for line in file:
            yield line.rstrip('\n')


def data_lines(path, size):
    lines = load_lines(path)
    seed(1)
    return sample(list(lines), size)
